//Game of Life with bidimensional rules
//Carlos Gershenson
//e-mail: carlos@jlagunez.iquimica.unam.mx

#include <graphics.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include "tools.h"

typedef unsigned char uchar;
typedef unsigned int uint;

const MAXX=64;
const MAXY=32;

typedef struct{
  float l;
  float d;
} bdlv;

bdlv m[MAXX][MAXY], n[MAXX+1][MAXY+1];
char g='',g2='';
int cyclic=1;       //flag for cyclic matrix
int proj=1;	    //flag for drawing mdl or it's projection in FL
long gen;	    //# of generations
int s1,s2,s3,s4;    //stats
int s12,s22,s32,s42;    //stats
int s13,s23,s33,s43;    //stats
int pause=0,stats=0;//flags
char filename[12];

void FP(uint color, uchar r,uchar g,uchar b){
//change palete colors
//courtesy of Prof. Robert G. Murcio
     asm{
       MOV AH,10H     //{asigna int}
       MOV AL,10H     //{asigna subfucion}
       MOV BX,color   //{color a cambiar}
       MOV CH,g       //{cantidad de rojo, verde y azul}
       MOV CL,b
       MOV DH,r
       INT 10H
     }
}

int col(int x){
  switch (x){
    case 6: return 20;
    case 8: return 56;
    case 9: return 57;
    case 10: return 58;
    case 11: return 59;
    case 12: return 60;
    case 13: return 61;
    case 14: return 62;
    case 15: return 63;
    default: return x;
  }
}

void iniPalette(){
  outtextxy(0,0,"Loading Color Palette...");
  for (int i=0;i<16;i++)
    FP (col(i),i,0,(i*4)+3);
  for (i=0;i<=MAXX;i++) {n[i][MAXY].l=0; n[i][MAXY].d=0;}
  for (i=0;i<=MAXY;i++) {n[MAXX][i].l=0; n[MAXX][i].d=0;}
}

int loadFile(char * name){
  FILE * f;
  char * ch;
  char c;
  char cx[3],cy[3],cl[10],cd[10];
  int flag,x,y;
  float v;
  long pos;
  for (int i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++){
      m[i][j].l=0;
      m[i][j].d=1;
    }
  if ((f = fopen(name, "r")) == NULL)
  {
     fprintf(stderr, "The file doesn't exist\n");
     getch();
     return(1);
  }
  pos=0;
  flag=1;
  while(flag){
    pos=0;
    fread(ch,1, 1, f);c= * ch;
    while((c!='\n')&&(flag)){
      if (c=='E') flag=0;
      cx[pos]=c;
      fread(ch,1, 1, f);c= * ch;
      pos++;
    }
    cx[pos]='\0';
    pos=0;
    fread(ch,1, 1, f);c= * ch;
    while((c!='\n')&&(flag)){
      if (c=='E') flag=0;
      cy[pos]=c;
      fread(ch,1, 1, f);c= * ch;
      pos++;
    }
    cy[pos]='\0';
    pos=0;
    fread(ch,1, 1, f);c= * ch;
    while((c!='\n')&&(flag)){
      if (c=='E') flag=0;
      cl[pos]=c;
      fread(ch,1, 1, f);c= * ch;
      pos++;
    }
    cl[pos]='\0';
    pos=0;
    fread(ch,1, 1, f);c= * ch;
    while((c!='\n')&&(flag)){
      if (c=='E') flag=0;
      cd[pos]=c;
      fread(ch,1, 1, f);c= * ch;
      pos++;
    }
    cd[pos]='\0';
    if (flag){
      m[atoi(cx)][atoi(cy)].l=atof(cl);
      m[atoi(cx)][atoi(cy)].d=atof(cd);
    }
    cout<<"x:"<<cx<<" y:"<<cy<<" v:"<<cl<<','<<cd<<'\n';
  }
  fclose(f);
  return(0);
}

float projl(bdlv x){ //projects life & death values into a fuzzy life value
  return (x.l-x.d+1)/2;
}


void paint (){
  BeforePaint();
  setcolor(15);
  line(0,320,640,320);
  outtextxy(0,330,"MDL Game of Life");
  outtextxy(0,340,"Gershenson Technologies Inc., 1999");
  outtextxy(390,330,"(M)DL/projection in FL  (S)tats");
  outtextxy(480,340,"(P)ause  Esc to exit");

  if (proj){
    for (int i=0; i<MAXX; i++)
      for (int j=0; j<MAXY; j++)
	if (projl(m[i][j])>0){
	  setcolor(projl(m[i][j])*15);
	  setfillstyle(1,projl(m[i][j])*15);
	  fillellipse((i*10)+5,(j*10)+5,4,4);
	}
  }
  else{
    for (int i=0; i<MAXX; i++)
      for (int j=0; j<MAXY; j++){
	if (m[i][j].l>0){
	  setcolor(m[i][j].l*15);
	  setfillstyle(1,m[i][j].l*15);
	  fillellipse((i*10)+3,(j*10)+5,2,4);
	}
	if (m[i][j].d>0){
	  setcolor(m[i][j].d*15);
	  setfillstyle(1,m[i][j].d*15);
	  fillellipse((i*10)+7,(j*10)+5,2,4);
	}
      }
  }
  AfterPaint();
  if (pause){
    getch();
    pause=0;
  }
}

float aroundl(int x, int y){   //sum of (if) cyclic live neighbor cells of x,y
  int x1,x2,y1,y2;
  float v;
  if ((cyclic)&&((x-1)<0)) x1=MAXX-1; else if ((x-1)<0)x1=MAXX;else x1=x-1;
  if ((cyclic)&&((x+1)>=MAXX)) x2=0; else if ((x+1)>=MAXX)x2=MAXX;else x2=x+1;
  if ((cyclic)&&((y-1)<0)) y1=MAXY-1; else if ((y-1)<0)y1=MAXY;else y1=y-1;
  if ((cyclic)&&((y+1)>=MAXY)) y2=0; else if ((y+1)>=MAXY)y2=MAXY;else y2=y+1;
  v=n[x1][y1].l+n[x1][y].l+n[x1][y2].l+n[x][y1].l+
    n[x][y2].l+n[x2][y1].l+n[x2][y].l+n[x2][y2].l;
  return v;
}

float aroundd(int x, int y){//sum of (if) cyclic dead neighbor cells of x,y
  int x1,x2,y1,y2;
  float v;
  if ((cyclic)&&((x-1)<0)) x1=MAXX-1; else if ((x-1)<0)x1=MAXX;else x1=x-1;
  if ((cyclic)&&((x+1)>=MAXX)) x2=0; else if ((x+1)>=MAXX)x2=MAXX;else x2=x+1;
  if ((cyclic)&&((y-1)<0)) y1=MAXY-1; else if ((y-1)<0)y1=MAXY;else y1=y-1;
  if ((cyclic)&&((y+1)>=MAXY)) y2=0; else if ((y+1)>=MAXY)y2=MAXY;else y2=y+1;
  v=n[x1][y1].d+n[x1][y].d+n[x1][y2].d+n[x][y1].d+
    n[x][y2].d+n[x2][y1].d+n[x2][y].d+n[x2][y2].d;
  return v;
}

void trans(){ //transition
  float v,w,x,y;
  s1=s2=s3=s4=0;
  s12=s22=s32=s42=0;
  s13=s23=s33=s43=0;
  for (int i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++)
      n[i][j]=m[i][j];
  for (i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++){
      v=aroundl(i,j);
      x=n[i][j].l;
      if (x>0){
	if ((v<=1)||(v>=4)) m[i][j].l=0;
	if ((1<v)&&(v<2)) m[i][j].l=v-1;
	if ((2<=v)&&(v<=3)) m[i][j].l=1;
	if ((3<v)&&(v<4)) m[i][j].l=4-v;
      }
      else{
	if ((2<v)&&(v<=3)) m[i][j].l=v-2;
	if ((3<v)&&(v<4)) m[i][j].l=4-v;
      }
      w=aroundd(i,j);
      y=n[i][j].l;
      if (y>0){
	if ((w<=4)||(w>=7)) m[i][j].d=1;
	if ((6<w)&&(w<7)) m[i][j].d=7-w;
	if ((5<=w)&&(w<=6)) m[i][j].d=0;
	if ((4<w)&&(w<5)) m[i][j].d=v-4;
      }
      else{
	if ((5<=w)&&(w<6)) m[i][j].d=6-w;
	if ((4<w)&&(w<5)) m[i][j].d=w-4;
      }
      if(m[i][j].l==1)
	s1++;
      else
	if(m[i][j].l==0)
	  s2++;
	else
	  if(m[i][j].l>=0.5)
	    s3++;
	  else
	    s4++;
      if(m[i][j].d==1)
	s12++;
      else
	if(m[i][j].d==0)
	  s22++;
	else
	  if(m[i][j].d>=0.5)
	    s32++;
	  else
	    s42++;
      if(projl(m[i][j])==1)
	s13++;
      else
	if(projl(m[i][j])==0)
	  s23++;
	else
	  if(projl(m[i][j])>=0.5)
	    s33++;
	  else
	    s43++;
  }
}

void rnd(){
  int pct=-1,pct1=-1;
  float t;
  cout<<"Around what percentage should be with life cells? (0-100)\n";
  while ((pct<0)||(pct>100)) cin>>pct;
  cout<<"Around what percentage of it should be with (1,0)? (0-100)\n";
  while ((pct1<0)||(pct1>100))cin>>pct1;
  for (int i=0; i<MAXX; i++)
    for (int j=0; j<MAXY; j++)
      if (pct>random(100))
	if (pct1>random(100)){
	  m[i][j].l=1;
	  m[i][j].d=0;
	}
	else{
	  m[i][j].l=random(101)*0.01;
	  m[i][j].d=random(101)*0.01;
	}
      else{
	m[i][j].l=0;
	m[i][j].d=1;
      }
}

void stts(){
  char cad[10];
  BeforePaint();
  setcolor(15);
  outtextxy(0,0,"MDL Life Stats");
  outtextxy(0,30,"Number of Generations:");
  itoa(gen, cad, 10);
  outtextxy(250,30,cad);
  outtextxy(0,50,"Number of life=1 :");
  itoa(s1, cad, 10);
  outtextxy(250,50,cad);
  outtextxy(0,60,"Number of life=0 :");
  itoa(s2, cad, 10);
  outtextxy(250,60,cad);
  outtextxy(0,70,"Number of 0.5<=life<1 :");
  itoa(s3, cad, 10);
  outtextxy(250,70,cad);
  outtextxy(0,80,"Number of 0<life<0.5 :");
  itoa(s4, cad, 10);
  outtextxy(250,80,cad);
  outtextxy(0,90,"Number of death=1 :");
  itoa(s12, cad, 10);
  outtextxy(250,90,cad);
  outtextxy(0,100,"Number of death=0 :");
  itoa(s22, cad, 10);
  outtextxy(250,100,cad);
  outtextxy(0,110,"Number of 0.5<=death<1 :");
  itoa(s32, cad, 10);
  outtextxy(250,110,cad);
  outtextxy(0,120,"Number of 0<death<0.5 :");
  itoa(s42, cad, 10);
  outtextxy(250,120,cad);
  outtextxy(0,130,"Number of projl=1 :");
  itoa(s13, cad, 10);
  outtextxy(250,130,cad);
  outtextxy(0,140,"Number of projl=0 :");
  itoa(s23, cad, 10);
  outtextxy(250,140,cad);
  outtextxy(0,150,"Number of 0.5<=projl<1 :");
  itoa(s33, cad, 10);
  outtextxy(250,150,cad);
  outtextxy(0,160,"Number of 0<projl<0.5 :");
  itoa(s43, cad, 10);
  outtextxy(250,160,cad);
  outtextxy(0,200,"Press any key...");
  AfterPaint();
  getch();
  stats=0;
}

int main(void){
   randomize();
   while (g!='0'){
     clrscr();
     textcolor(9);
     cout<<"Wellcome to MDL Life!\n"<<'\n';
     cout<<"1) Load from file\n";
     cout<<"2) Random\n";
     cout<<"3) Cyclic";if (cyclic) cout<<" (On)\n";else cout<<" (Off)\n";
     cout<<"0) Exit\n";
     g=getch();
     if (g==27) exit(0);
     switch (g){
       case'1':cout<<"Please enter the file name\n";
	       cin>>filename;
	       if (loadFile(filename))
		 g='';
	       break;
       case'2':rnd();break;
       case'3':if (cyclic)
		 cyclic=0;
	       else
		 cyclic=1;
	       break;
     }
     if ((g=='1')||(g=='2')){
       IniGraphics();
       iniPalette();
       g2='';
       gen=0;
       while (g2!=27){
	 while (!kbhit()){
	   if (stats) stts();
	   paint();
	   trans();
	   gen++;
	 }
	 g2=getch();
	 switch (g2){
	   case 'p':pause=1;break;
	   case 's':stats=1;break;
	   case 'm':if (proj) proj=0; else proj=1;break;
	 }
       }
       closegraph();
     }
   }
   return 0;
}
